
import os

import amrs.vis

import matplotlib.patheffects as PathEffects
import cartopy.crs as ccrs
import numpy as np
import pandas


# point sources in SMARTCARB model domain
def read_point_sources():
    return pandas.read_csv(
        os.path.join(os.path.dirname(__file__), '..', 'files', 'sources.csv'), 
        index_col=0, names=['UTF-8', 'longitude', 'latitude', 'type']
    ).to_xarray()
    

def add_hot_spots(ax, color='black', size='medium', suffixes=None, names=None,
                  va='flex', ha='left', fontdict=None, bbox=None, ms=None,
                  do_path_effect=False, mec=None, add_labels=True):
    """

    """
    point_sources = read_point_sources()
    
    if names is None:
        names = sorted(point_sources.ASCII.values)

        
    if add_labels:
        ax.plot(0, -90, 'D', color=color, transform=ccrs.PlateCarree(), mec=mec,
                ms=ms, label='city', zorder=0)
        ax.plot(0, -90, 'o', color=color, transform=ccrs.PlateCarree(), mec=mec,
                ms=ms, label='power plant', zorder=0)
    
        
        
    for i, key in enumerate(names, 1):
        
        source = point_sources.sel(index=key)
        lon = float(source['longitude'])
        lat = float(source['latitude'])
        text = str(source['UTF-8'].values)
        type_ = str(source['type'].values)
        
        marker = 'D' if type_ == 'city' else 'o'
        
        ax.plot(lon, lat, marker, color=color, transform=ccrs.PlateCarree(), mec=mec,
                ms=ms, label='_none_', zorder=10)

        if isinstance(ha, str):
            align = ha
        else:
            align = ha[i]
            
        if key in ['Staudinger', 'Heyden', 'Melnik']:
            align = 'left'
            
        if key == 'Pocerady':
            va = 'bottom'
        elif key == 'Prunerov':
            va = 'top'
        else:
            va = 'center'
            
        t = ax.text(lon, lat, f'  {text}  ', va=va, ha=align, fontsize=size,
                    transform=ccrs.PlateCarree(), color=color, clip_on=True,
                    fontdict=fontdict, bbox=bbox)
        t.set_clip_on(True)
        if do_path_effect:
            t.set_path_effects([PathEffects.withStroke(linewidth=2.5, foreground='w')])




def visualize(data, do_zoom=True, show_clouds=True, alpha=None,
              is_sentinel_5=False, data7=None, vlim=(-2,2),
              legend_loc='upper right', simple_legend=True, add_multiple_sources=True,
              tracer=None, show_true_plumes=True,
              source=None, names=None, figwidth=6.97, vmin=None, vmax=None, add_wind=True,
              zoom_on=None, min_height=None, ax=None, cax=None, cmap=None,
              marker='+', markersize=3, domain=None, label=None):
    """
    Visualize detected plume.

    do_zoom :: zoom on plume instead of showing full domain
    show_clouds :: if true overlay cloudy pixels
    """
    if ax is None and cax is None:
        fig = None
    else:
        fig = ax, cax
        
    if tracer is None:
        threshold = {'CO2': 0.01, 'NO2': 0.30}[tracer[:3]]
        tracer = data.tracer
        clouds = data.is_cloudy if show_clouds else None
        clouds.attrs['threshold'] = threshold

    else:
        if show_clouds:
            threshold = {'CO2': 0.01, 'NO2': 0.30}[tracer[:3]]
            clouds = data.clouds >= threshold
            clouds.attrs['threshold'] = threshold
        else:
            clouds = None
    
    if tracer[:3] == 'CO2':

        # compute vmin and vmax
        if vmin is None and vmax is None:
            mean_bg = np.nanmean(data.CO2)
            vmin = round(mean_bg + vlim[0])
            vmax = round(mean_bg + vlim[1])

        fig = amrs.vis.make_level2_map(
            data.lonc, data.latc, data[tracer], vmin=vmin,
            vmax=vmax, label='XCO$_2$ [ppm]' if label is None else label,
            clct=clouds, alpha=alpha,
            domain=domain, fig_size=figwidth, fig=fig, cmap=cmap
        )
    else:
        if vmin is None:
            vmin = 0.0
        if vmax is None:
            vmax = 10.0
            
        fig = amrs.vis.make_level2_map(
            data.lonc, data.latc, data.NO2 / 1e15, vmin=vmin, vmax=vmax, truncate_cmap=True,
            label='NO$_2$ column [10$^{15}$ molecules per cm$^2$]' if label is None else label,
            clct=clouds, alpha=alpha,
            domain=domain, fig_size=figwidth, fig=fig, cmap=cmap
        )
        
    if ax is None:
        ax, cax = fig.get_axes()
    
    if isinstance(source, str):
        source = [source]
    
    # use names for map labels (if provided) but plumes are plotted based on data.source
    if source is None:
        try:
            source = data['source'].values
        except KeyError:
            source = []
            
        if np.ndim(source) == 0:
            source = [str(source)]
            
        
    if names is None:
        names = [o for o in source]
   
    if add_wind:
        add_wind_marker(ax, data, sources=source)
    
    # FIXME
    ms = 0.0 if add_wind else 5.0
    add_hot_spots(ax, color='black', mec='w', ms=ms, ha='right', size='small', names=names,
                  do_path_effect=True)

    # lines and labels
    if show_clouds:
        label = 'Cloud fraction > %d%%' % (100*clouds.threshold)
        ax.plot(domain.pollon, domain.pollat, marker='s', mec='k', mfc='w', ls='',
                transform=ccrs.PlateCarree(), label=label)

    # true plumes
    if show_true_plumes:
        for tracer, ls in [('CO2', '-'), ('NO2', '--')]:
            name = '%s_true_plume' % tracer

            if name not in data:
                continue

            if tracer == 'CO2':
                if is_sentinel_5:
                    thr = data7[name].attrs['signal threshold (ppm)']
                else:
                    thr = data[name].attrs['signal threshold (ppm)']
                label = 'CO$_2$ plume (>%.2f ppm)' % thr

            elif tracer == 'NO2':
                try:
                    thr = data[name].attrs['signal threshold (molecules cm-2)']
                    label = r'NO$_2$ plume (>%.1f$\times$10$^{15}$ molec. cm$^{-2}$)' % (thr / 1e15)
                except KeyError:
                    thr = None
                    label = None
            else:
                label = name

            if hasattr(data, name):
                tolerance = 0.5 if is_sentinel_5 else 1.0
                amrs.vis.draw_contours(ax, data.lon, data.lat,
                                       data[name], label=label, ls=ls,
                                       tolerance=tolerance)
                continue

            if hasattr(data7, name):
                amrs.vis.draw_contours(ax, data7.lon, data7.lat,
                                       data7[name], label=label, ls=ls)

    # detected plumes
    if ('detected_plume' in data and data.detected_plume is not None and
        np.sum(data.detected_plume) > 0):
        
        if np.ndim(data.detected_plume) == 3:
            plume = data.detected_plume.values.astype(bool)
        else:
            plume = data.detected_plume.values[:,:,np.newaxis].astype(bool)

        # plot multiple detections
        multiple_sources = plume.sum(2) > 1
        
        if add_multiple_sources:
            lon0 = data.lon.values[multiple_sources]
            lat0 = data.lat.values[multiple_sources]
            n_pixels = np.sum(multiple_sources)

            ax.plot(lon0, lat0, marker=marker, color='r', alpha=0.5,
                    ms=markersize, ls='', transform=ccrs.PlateCarree(),
                    label='Multiple sources (%d px)' % n_pixels)
        
        # plot other plumes
        if zoom_on is not None:
            other_sources = data.detected_plume.any('source') & ~data.detected_plume.sel(source=zoom_on)
            lon0 = data.lon.values[other_sources]
            lat0 = data.lat.values[other_sources]

            ax.plot(lon0, lat0, marker=marker, color='r', alpha=0.5,
                    ms=markersize, ls='', transform=ccrs.PlateCarree(),
                    label='Other sources')
        
        
        # iterate over several plume detections
        for j in range(plume.shape[2]):

            if zoom_on is not None and zoom_on != source[j]:
                continue
                
            name = cfg.KEY2UNICODE[source[j]]

            lon0 = data.lon.values[plume[:,:,j] & ~multiple_sources]
            lat0 = data.lat.values[plume[:,:,j] & ~multiple_sources]
            n_pixels = lon0.size

            if source is None and n_pixels == 0:
                continue

            if simple_legend:
                label = '%s (%d px)' % (name, n_pixels)
            else:
                label = 'detected plume (q = %.3f, n = %d)\nwith %d pixels' 
                label %= (
                    data.attrs['probability for z-value'],
                    data.attrs['size of neighborhood'],
                    n_pixels
                )
            
            ax.plot(lon0, lat0, marker=marker, mec='k',
                    ms=2*markersize if is_sentinel_5 else markersize,
                    mfc='k', ls='',
                    transform=ccrs.PlateCarree(), label=label,
                    alpha=0.5 if zoom_on is None else 1.0)

        
    if do_zoom:
        
        if zoom_on is not None:
            data = data.sel(source=zoom_on)
        
        if np.ndim(data['detected_plume']) == 3:
            lon_o = data['lon_o'].values
            lat_o = data['lat_o'].values
            
            lon = np.concatenate([lon_o, data.lon.values[data['detected_plume'].any('source')].flatten()])
            lat = np.concatenate([lat_o, data.lat.values[data['detected_plume'].any('source')].flatten()])
        else:
            if 'lon_o' in data:
                lon_o = data['lon_o'].values
                lat_o = data['lat_o'].values
            else:
                lon_o, lat_o, _ = misc.get_source_origin(source[0])
                
            lon = np.concatenate([[lon_o], data.lon.values[data['detected_plume']].flatten()])
            lat = np.concatenate([[lat_o], data.lat.values[data['detected_plume']].flatten()])

            
        rlon_0, rlat_0 = amrs.misc.transform.rotpole2wgs(lon, lat, cfg.pollon,
                                                         cfg.pollat, inverse=True)
        dy = 0.1
        dx = 0.1
        
        xmin, xmax = min(rlon_0) - dx, max(rlon_0) + dx
        ymin, ymax = min(rlat_0) - dy, max(rlat_0) + dy
        
        # make figure at least 200 km high
        if min_height is None:
            min_height = 2.0 if zoom_on is None else 1.0
            
        if ymax - ymin < min_height:
            ymid = ymin + 0.5 * (ymax - ymin)
            ymin = ymid - 0.5 * min_height
            ymax = ymid + 0.5 * min_height
   
        # make figure have same aspect ratio as model domain
        dx = 0.5 * (1.15 * (ymax - ymin) - (xmax - xmin))
        xmin = xmin - dx
        xmax = xmax + dx
        
        # shift limits when outside of domain boundary
        if xmin < domain.rlon[0]:
            shift = domain.rlon[0] - xmin
            xmin += shift
            xmax += shift
            
        if xmax > domain.rlon[-1]:
            shift = xmax - domain.rlon[-1]
            xmin -= shift
            xmax -= shift
        
        ax.set_xlim(xmin, xmax)
        ax.set_ylim(ymin, ymax)
        
    else:
        xmin, xmax = ax.get_xlim()
        ymin, ymax = ax.get_ylim()
        ax.set_xlim(xmin+0.01*50, xmax-0.01*50)
        ax.set_ylim(ymin+0.01*50, ymax-0.01*50)

    legend = ax.legend(loc=legend_loc, numpoints=1, markerscale=1.5,
                       fontsize='small')

    for lh in legend.legendHandles: 
        lh._legmarker.set_alpha(1)

    return fig


